// Einfache Musikbox in C++, welcher in Textdateien (*.MUS) gespeicherte 
// Lieder abspielt
// (c) 1996 by Andreas Meile, HTL Brugg-Windisch
// Anwendung der Formel f=440*2^(n/12) fr die Berechnung der Frequenzen

#include <stdio.h>
#include <dos.h>  // enthlt Zeitverzgerungsfunktion
#include <pc.h>   // enthlt Tonerzeugbefehl
#include <math.h> // wegen pow()-Funktion

// Auswerten einer Note
// Format: {A|B|...|H|a|b|c|...|h|X} (Notenname)
// [#|b] (Erhhung/Erniedriegung um 1)
// ['|"] (Oktave: eingestrichen/zweigestrichen)
// ^letzte Zeichen drfen hintereinander mehrfach vorkommen
// z.B. F#''' ist auch erlaubt
// [.|_] Artikulation: . = 
// Dauer, wahlweise im Format X oder X/X (z.B. fr Achtel- und Sechzentel-
// noten sowie Triolen) in Anzahl Schlge

const int paus_wert = 32767; // Wert fr fehlerhafte Note

const int staccato = -1; // Kennung fr Artikulation "staccato"
const int normal = 0;    // normal
const int tenuto = 1;    // Tenuto
const int legato = 2;    // gebunden

const float ton_a = 440.0; // Grundkammerton a = 440 Hz

const int tempo_minute = 60 * 1000; // Tempokonstante 1 min = 60'000 ms
// d.h. um zwischen Anzahl Schlge pro Minute und Pausenwert bei delay()
// umrechnen zu knnen

// Notenwerte n
const int not_tab[9] = {
   0,  // Kammerton a
   1,  // Kammerton b
  -9,  // Kammerton c
  -7,  // Kammerton d
  -5,  // Kammerton e
  -4,  // Kammerton f
  -2,  // Kammerton g
   2,  // Kammerton h
};

// Speicher fr Musikstck

const int max_laenge = 1000; // max. 1000 Noten erlaubt

int frequenz[max_laenge], // Notenfrequenzen
    dauer[max_laenge],    // Dauer der Note (ms)
    pause[max_laenge],    // Pause dazwischen

    groesse;              // Grsse des Stckes

// Division mit korrekter Rundung
int div_rund(int z, int n) {
  if(n == 0) {
    printf("Fehler");
    exit(1);
  }
  return (z + (n >> 1)) / n;
}

// Berechnung der Tonfrequenz
float ton_frequenz(int note) {
  return ton_a * pow(2.0, float(note) / 12.0);
}

// Einlesen des Musikstckes
void lies_musikstueck_ein(char *d_name) {
  FILE *d_handle;   // Datei-Handle

  int ch,           // einzulesendes Zeichen
      n_wert,       // Notenwert
      artik,        // Art der Artikulation
      dau_z,        // Zhler der Notendauer
      dau_n,        // Nenner der Notendauer
      n_ges_laenge, // Gesamtlnge einer Note
      tau,          // charakteristischer Lckenlngenfaktor fr Artikulation      
      tempo = 120;  // Momentanes Tempo     

  d_handle = fopen(d_name, "r");
  if(d_handle == NULL) {
    printf("Fehler beim ffnen der Datei %s\n", d_name);
    exit(3);
  }

  // Einlesen : Read-Ahead-Technik nach Jakson verwenden

  groesse = 0; // Leeres Stck zu Beginn

  ch = fgetc(d_handle);
  while(ch != EOF) {
    switch (ch) {
      case '%' : // Kommentarzeile
        // => diesen berlesen
        while(ch != EOF && ch != '\n')
          ch = fgetc(d_handle);
        if(ch != EOF) // Line-Terminator berspringen
          ch = fgetc(d_handle);
        break;
      case ' '  : // Leerzeichen
      case '\n' : // Zeilenende
      case '\t' : // Tabulatorzeichen
        ch = fgetc(d_handle);  // Zeichen, welches bersprungen werden
        // soll
        break;
      case 'T' : // Angabe des Tempo
        ch = fgetc(d_handle);
        if(ch != '=') {
          printf("Tempoangabe muss im Format T=Wert erfolgen\n");
          fclose(d_handle);
          exit(1);
        }
        ch = fgetc(d_handle);
        tempo = 0;
        while(ch >= '0' && ch <= '9') {
          tempo = 10 * tempo + ch - '0';
          ch = fgetc(d_handle);
        }
        if(tempo == 0) {
          printf("Ungltiger Wert fr Tempoangabe gefunden\n");
          fclose(d_handle);
          exit(1);
        }
        break;
      default :  // es muss ein gltiger Eintrag vorliegen
        if(ch == 'X') // Pause
          n_wert = paus_wert;
        else if(ch >= 'A' && ch <= 'H')   // Bassnote
          n_wert = not_tab[ch - 'A'] - 12;
        else if(ch >= 'a' && ch <= 'h') // Melodienote
          n_wert = not_tab[ch - 'a'];
        else { // Fehlerhafte Note
          printf("Fehler beim Einlesen: Ungltige Note gefunden\n");
          fclose(d_handle);
          exit(1);
        }
        ch = getc(d_handle);  // Nchstes Zeichen

        // Halbtonzeichen und Oktave auswerten
        while(ch == '#' || ch == 'b' || ch == '\'' || ch == '\"') {
          if(n_wert == paus_wert) {
            printf("Oktav- und Vorzeichenangabe bei Pausen nicht zulssig\n");
            fclose(d_handle);
            exit(1);
          }
          switch(ch) {
            case '#' :  // Erhhung um einen Halbton (-is, z.B. Fis)
              n_wert++;
              break;
            case 'b' :  // Erniedrigung um einen Halbton (-es, z.B. Des)
              n_wert--;
              break;
            case '\'' : // Ein Strich mehr
              if(n_wert < -9) // Bassnote, d.h. H oder tiefer
                n_wert -= 12;
              else  // Melodienote, d.h. c oder hher
                n_wert += 12;
              break;
            default : // Zwei Striche mehr
              if(n_wert < -9) // Bassnote, d.h. H oder tiefer
                n_wert -= 24;
              else  // Melodienote, d.h. c oder hher
                n_wert += 24;
          }
          ch = fgetc(d_handle);
        }

        // Artikulationen auswerten
        if(ch == '.' || ch == '-' || ch == '_') {
          if(n_wert == paus_wert) {
            printf("Artikulationszeichen bei Pausen nicht zulssig\n");
            fclose(d_handle);
            exit(1);
          }
          switch(ch) {
            case '.' : // Staccato
              artik = staccato;
              break;
            case '-' : // Tenuto
              artik = tenuto;
              break;
            default : // Legato
              artik = legato;
          }
          ch = fgetc(d_handle);
        } else // Standardlnge
          artik = normal;

        // Tondauer auswerten
        dau_z = 0;
        while(ch >= '0' && ch <= '9') {
          dau_z = 10 * dau_z + ch - '0';
          ch = fgetc(d_handle);
        }
        if(dau_z == 0) {
          printf("Fehler: positive Notendauer erwartet\n");
          fclose(d_handle);
          exit(1);
        }

        if(ch == '/') { // Noch Nenner vorliegend
          ch = fgetc(d_handle); // Bruchstrich berlesen
          dau_n = 0;
          while(ch >= '0' && ch <= '9') {
            dau_n = 10 * dau_n + ch - '0';
            ch = fgetc(d_handle);
          }
          if(dau_n == 0) {
            printf("Bei Notendauerangabe mit Bruch sollte ein positiver "
            "Nenner vorliegen\n");
            fclose(d_handle);
            exit(1);
          }
        } else
          dau_n = 1;  // Standardwert fr Nenner

        // Vollstndige Verarbeitung der Note
        n_ges_laenge = div_rund(tempo_minute * dau_z,
        tempo * dau_n); // Gesamttaktlnge

        if(n_wert == paus_wert) {
          dauer[groesse] = -1; // Kennzeichen fr Pause
          pause[groesse] = n_ges_laenge; // Gesamte Lnge als Pause
        } else {
          frequenz[groesse] = int(ton_frequenz(n_wert) + 0.5);
          // Frequenz gemss der Formel fr die temp. Stimmung von J.S.Bach
          if(artik == legato) {
            dauer[groesse] = n_ges_laenge; // Gesamte Lnge als Ton
            pause[groesse] = -1; // Kennzeichen fr nahtlos gebundene Tne
          } else {
            switch(artik) {
              case tenuto :
                tau = 1;
                break;
              case normal :
                tau = 3;
                break;
              default : // Staccato
                tau = 7;
            }
            if(4 * dau_z < dau_n) // Kurzer Ton
              pause[groesse] = div_rund(n_ges_laenge * tau, 8);
            else
              pause[groesse] = div_rund(tau * tempo_minute, 32 * tempo);
            dauer[groesse] = n_ges_laenge - pause[groesse];
          }
        }
        groesse++;
    }
  }
  fclose(d_handle);
}

// Musikstck abspielen
void spiele_musikstueck() {
  int i; // Zhler

  for(i = 0; i < groesse; i++) {
    if(dauer[i] > 0) {     // Falls keine Pause
      sound(frequenz[i]);  // Ton spielen
      delay(dauer[i]);     // Tondauer
    }
    if(pause[i] > 0) {     // Falls kein Legato
      nosound();           // Abbrechen
      delay(pause[i]);        // Pause
    }
  }
}

void main() {
  char d_name[80]; // fr Dateiname

  printf("\e[0m\e[2J\e[2;1HPC-Musikbox\n");
  printf("(c) 1996 by Andreas Meile, HTL Brugg-Windisch\n\n");

  printf("Name der abzuspielenden Musikdatei? ");
  gets(d_name);

  lies_musikstueck_ein(d_name); // Stck einlesen
  spiele_musikstueck();         // und abspielen
}
